use crate::config::config::Config;
use crate::store::{
    nsobject::NsObject,
    object::{Obj, ObjStatus},
};
use crate::utils::path;
use std::collections::HashMap;
use std::fs;
use std::path::{Path, PathBuf};
use std::sync::{Arc, Mutex};
use threadpool::ThreadPool;

pub struct StatusArgs {
    pub nostyle: bool,
}

struct ObjStatuses {
    created: Arc<Mutex<HashMap<String, Obj>>>,
    modified: Arc<Mutex<HashMap<String, Obj>>>,
    deleted: Arc<Mutex<HashMap<String, Obj>>>,
}

impl ObjStatuses {
    fn new() -> Self {
        ObjStatuses {
            created: Arc::new(Mutex::new(HashMap::new())),
            modified: Arc::new(Mutex::new(HashMap::new())),
            deleted: Arc::new(Mutex::new(HashMap::new())),
        }
    }

    fn push_created(&self, key: String, value: Obj) {
        self.created.lock().unwrap().insert(key, value);
    }

    fn push_modified(&self, key: String, value: Obj) {
        self.modified.lock().unwrap().insert(key, value);
    }

    fn push_deleted(&self, key: String, value: Obj) {
        self.deleted.lock().unwrap().insert(key, value);
    }
}

pub fn exec(args: StatusArgs, config: Config) {
    // use root of repo if no custom path has been set by the command
    let root = if config.is_custom_execution_path {
        config.execution_path.clone()
    } else {
        config.get_root_unsafe().to_path_buf()
    };

    let pool = ThreadPool::new(4);
    let repo_root = config.get_root_unsafe().clone();
    let res = Arc::new(ObjStatuses::new());

    let pool_clone = pool.clone();
    let res_clone = Arc::clone(&res);
    pool.execute(move || {
        compare_dir(pool_clone, &repo_root, &root, res_clone);
    });

    pool.join();

    for entry in res.created.lock().unwrap().iter() {
        println!("created: {}", entry.0);
    }
    for entry in res.modified.lock().unwrap().iter() {
        println!("modified: {}", entry.0);
    }
    for entry in res.deleted.lock().unwrap().iter() {
        println!("deleted: {}", entry.0);
    }
    // find moved and copied
}

///
///
/// * `pool`:
/// * `root_path`: path of the repo's root
/// * `path`: path we should analyze
/// * `res`: the struct in which we should store the response
fn compare_dir(pool: ThreadPool, root_path: &PathBuf, path: &PathBuf, res: Arc<ObjStatuses>) {
    let entries = match fs::read_dir(path) {
        Ok(entries) => entries,
        Err(err) => {
            eprintln!("Failed to read {} ({err})", path.display());
            return;
        }
    };

    let mut local_files = HashMap::new();
    let mut local_dirs = HashMap::new();

    // Read locals files and folders
    for entry in entries {
        let entry = match entry {
            Ok(entry) => entry,
            Err(err) => {
                eprintln!("Failed to read entry {err}");
                continue;
            }
        };

        let repo_relative_entry = path::to_repo_relative(&entry.path(), root_path);
        let local_obj = Obj::from_local_path(&repo_relative_entry);

            if entry.path().is_dir() {
            if entry.path().ends_with(".nextsync") {
                continue;
            }

            if local_obj.is_new() {
                // TODO: opti move files in new directory
                res.push_created(local_obj.cpy_path(), local_obj);
            } else {
                local_dirs.insert(local_obj.cpy_path(), local_obj);

                // Analyze sub folder
                let pool_clone = pool.clone();
                let root_path_clone = root_path.clone();
                let res_clone = Arc::clone(&res);
                pool.execute(move || {
                    compare_dir(pool_clone, &root_path_clone, &entry.path(), res_clone);
                });
            }
        } else {
            if local_obj.is_new() {
                res.push_created(local_obj.cpy_path(), local_obj);
            } else {
                local_files.insert(local_obj.cpy_path(), local_obj);
            }
        }
    }

    // Read ns objects to find deleted
    let entries = NsObject::from_local_path(&path);
    for entry in entries.iter() {
        if entry.is_file() {
            match local_files.get(entry.get_obj_path().to_str().unwrap()) {
                None => res.push_deleted(path::to_string(entry.get_obj_path()), {
                    let mut obj = entry.get_local_obj();
                    obj.set_status(ObjStatus::Deleted);
                    obj
                }),
                Some(local_obj) if local_obj != entry => {
                    res.push_modified(local_obj.cpy_path(), local_obj.clone())
                }
                Some(_) => (),
            }
        } else {
            match local_dirs.get(entry.get_obj_path().to_str().unwrap()) {
                None => res.push_deleted(path::to_string(entry.get_obj_path()), {
                    let mut obj = entry.get_local_obj();
                    obj.set_status(ObjStatus::Deleted);
                    obj
                }),
                Some(_) => (),
            }
        }
    }
}
