use crate::store::object::ObjType;
use std::fs::File;
use std::io::{self, Read, Seek, SeekFrom, Write};
use std::path::PathBuf;

fn get_level(previous_level: u16, path: &PathBuf) -> (u16, IndexLevel) {
    let mut level = 0;
    let mut path = path.clone();
    while path.pop() {
        level += 1;
    }

    let index_level = if previous_level < level {
        IndexLevel::Down
    } else {
        IndexLevel::Up
    };

    (level, index_level)
}

enum IndexLevel {
    Up,
    Down,
}

struct IndexedObj {
    obj_type: ObjType,
    level: IndexLevel,
    path: PathBuf,
}

pub struct Indexer {
    index_file: PathBuf,
    previous_level: u16,
    indexed_objs: Vec<IndexedObj>,
}

impl Indexer {
    pub fn new(repo_root: &PathBuf) -> Self {
        let mut index_file = repo_root.clone();
        index_file.push(".nextsync");
        index_file.push("index");

        Indexer {
            index_file,
            previous_level: 0,
            indexed_objs: Vec::new(),
        }
    }

    pub fn load(&self) -> io::Result<()> {
        let mut file = File::open(&self.index_file)?;
        let mut str = String::new();

        // Skip reserved bytes
        let mut byte = [0; 1];
        file.read_exact(&mut byte)?;
        file.seek(SeekFrom::Start(1))?;

        // Read usize value
        let mut usize_bytes = [0; std::mem::size_of::<usize>()];
        file.read_exact(&mut usize_bytes)?;
        dbg!(ObjType::try_from(byte[0]));
        let usize_value = usize::from_le_bytes(usize_bytes);

        // Read PathBuf as string
        let mut buffer = Vec::new();
        file.read_to_end(&mut buffer)?;
        let path_str = String::from_utf8(buffer).unwrap();
        let path = PathBuf::from(path_str);

        println!("usize value: {}", usize_value);
        println!("Path: {:?}", path);
        Ok(())
    }

    fn index_obj(&mut self, path: PathBuf, obj_type: ObjType) {
        let (level, index_level) = get_level(self.previous_level, &path);
        self.previous_level = level;

        self.indexed_objs.push(IndexedObj {
            obj_type,
            level: index_level,
            path: path.clone(),
        });
    }

    pub fn index_file(&mut self, path: PathBuf) {
        self.index_obj(path, ObjType::Blob);
    }

    pub fn index_dir(&mut self, path: PathBuf) {
        self.index_obj(path, ObjType::Tree);
    }

    pub fn save(&self) -> io::Result<()> {
        let mut file = File::create(&self.index_file)?;

        // Write reserved bytes
        let variant = ObjType::Blob;
        let byte: u8 = variant.into();
        file.write_all(&[byte])?;

        // Write usize value
        let usize_value: usize = 12;
        let usize_bytes = usize_value.to_le_bytes();
        file.write_all(&usize_bytes)?;

        // Write PathBuf as string
        let path = PathBuf::from("/jodi/");
        let path_str = path.to_str().unwrap();
        file.write_all(path_str.as_bytes())?;

        Ok(())
    }
}
